"""
Provide mixins for throttling/rate limiting in views.

Depends on ``django-axes``.

.. todo:: Decouple from django-axes - make IP address getter function configurable.
"""

import warnings
from collections.abc import Container
from time import time
from typing import Literal

from django.core.cache import caches
from django.core.cache.backends.base import BaseCache
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import HttpRequest, HttpResponse, HttpResponseBase

from axes.helpers import get_client_ip_address

ONE_MINUTE = 60
ONE_HOUR = ONE_MINUTE * 60


class ThrottleMixin:
    """
    A very simple throttling implementation with, hopefully, sane defaults.

    You can specifiy the amount of visits (``throttle_visits``) a view can get,
    for a specific period (in seconds) ``throttle_period``.
    """

    throttle_visits = 100
    """
    Number of allowed visits in the specified period.
    """

    throttle_period = ONE_HOUR
    """
    Period/time window (in seconds) in which the visits are counted.

    Visits older than this window are discarded.
    """

    throttle_name = "default"
    """
    Identifier for the throttle, used in the cache key.
    """

    throttle_cache = "default"
    """
    Name of the cache (in ``settings.CACHES``) to use to track visits.

    .. note:: Ensure you use a globally shared cached. Local memory caches are limited
       to their respective Python process and not aware of other processes/caches.
    """

    throttle_403 = False
    """
    Marker to opt-in to return 403 responses.

    DeprecationWarning - implement :meth:`ThrottleMixin.handle_rate_limit_exceeded`
    instead or use the default 429 response.

    .. versionchanged:: 0.7.0

        The default is changed to return 429 instead of 403 and the attribute has been
        deprecated.
    """

    # get and options should always be fast. By default
    # do not throttle them.
    throttle_methods: Container[str] | Literal["all"] = (
        "post",
        "put",
        "patch",
        "delete",
        "head",
        "trace",
    )

    request: HttpRequest

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.throttle_403:
            warnings.warn(
                f"'{cls.__name__}.throttle_403' attribute is deprecated - consider "
                "returning a HTTP 429 response instead.",
                category=DeprecationWarning,
                stacklevel=2,
            )

    def get_throttle_cache(self) -> BaseCache:
        return caches[self.throttle_cache]

    def get_throttle_identifier(self) -> str:
        user = self.request.user
        return str(user.pk)

    def _get_throttle_window(self):
        """
        Calculate the start of the window based on the current time.

        This uses the current time (unix timestamp) as input and looks up the most
        recent moment when a non-completed block of ``trottle_period`` started. It is
        used as input in the cache key, meaning that once ``throttle_period`` has
        elapsed, the throttle quota is fully reinstated.

        Effectively, throttle intervals are fixed and not a sliding window.
        """
        current_time = int(time())
        return current_time - (current_time % self.throttle_period)

    def _get_num_visits_in_window(self) -> int:
        cache = self.get_throttle_cache()
        cache_key = (
            f"throttling_{self.get_throttle_identifier()}_"
            f"{self.throttle_name}_{self._get_throttle_window()}"
        )

        added = cache.add(cache_key, value=1, timeout=self.throttle_period)
        if (
            added
        ):  # key added, we had no counter before -> one visit returned and stored
            return 1

        try:
            return cache.incr(cache_key)
        except ValueError:  # XXX: when does this happen?
            return 1

    def should_be_throttled(self) -> bool:
        """
        Determine if throttling is enabled for the request.
        """
        if self.throttle_methods == "all":
            return True
        assert isinstance(self.request.method, str)
        return self.request.method.lower() in self.throttle_methods

    def check_rate_limit_exceeded(self) -> bool:
        """
        Determine if the rate limit is exceeded or not.

        The limit is considered exceeded when:

        * the request matches the conditions to be throttled
        * the amount of visits in the time window exceeds the maximum allowed
        """
        enabled = self.should_be_throttled()
        # deliberate method call after the *and* to benefit from short-circuiting and
        # avoid hitting the cache if it's not needed
        return enabled and self._get_num_visits_in_window() > self.throttle_visits

    def handle_rate_limit_exceeded(self) -> HttpResponseBase:
        """
        Return the appropriate response for throttled requests.

        Override this to customize behaviour. By default, an HTTP 429 response is
        returned.
        """
        if self.throttle_403:
            raise PermissionDenied()
        return HttpResponse("rate limit exceeded", status=429)

    def dispatch(self, request, *args, **kwargs):
        if self.check_rate_limit_exceeded():
            return self.handle_rate_limit_exceeded()
        return super().dispatch(request, *args, **kwargs)  # pyright:ignore[reportAttributeAccessIssue]


class IPThrottleMixin(ThrottleMixin):
    """
    Same behavior as ThrottleMixin except it limits the amount of tries
    per IP-address a user can access a certain view.
    """

    def get_throttle_identifier(self):
        ip_address = get_client_ip_address(self.request)
        if not ip_address:
            raise ImproperlyConfigured(
                "Could not determine IP address. "
                "Check your reverse proxy configuration."
            )
        return ip_address



from django.conf import settings as django_settings


def settings(request):
    """ """
    public_settings = (
        "GOOGLE_ANALYTICS_ID",
        "PROJECT_NAME",
        "RELEASE",
        "GIT_SHA",
    )

    context = {
        "settings": {k: getattr(django_settings, k, None) for k in public_settings},
    }

    if hasattr(django_settings, "SENTRY_CONFIG"):
        context.update(dsn=django_settings.SENTRY_CONFIG.get("public_dsn", ""))

    return context



"""
Provide hooks for drf-spectacular schema post-processing.

Some generated schemas by drf-spectacular contain issues that don't pass linter
validation or emit warnings. The hooks in this module address these.
"""

import itertools
from collections.abc import MutableMapping, Sequence


def remove_invalid_url_defaults(result, *args, **kwargs):
    """
    Fix ``URLField(default="")`` schema generation.

    An empty string does not satisfy the `format: uri` validation, and schema validation
    tools can trip on this.

    The majority of the code here is inspired by the built in
    :func:`drf_spectacular.hooks.postprocess_schema_enums` hook.

    .. todo:: contribute a fix to upstream library.
    """

    def _is_type(schema_type: Sequence, expected: str) -> bool:
        if schema_type == expected:
            return True
        return not isinstance(schema_type, str) and expected in schema_type

    def iter_parameter_schemas():
        for path in result.get("paths", {}).values():
            for operation in path.values():
                if not (parameters := operation.get("parameters")):
                    continue
                for parameter in parameters:
                    yield parameter["schema"]

    def iter_component_schemas(schema):
        """
        Walk the schema definitions recursively so that each schema can be processed.
        """
        match schema:
            # array schema type variants
            case {"type": Sequence() as types} if _is_type(types, "array"):
                if item_schema := schema.get("items"):
                    yield item_schema
                    yield from iter_component_schemas(item_schema)

                if prefix_items := schema.get("prefixItems"):
                    for item_schema in prefix_items:
                        yield item_schema
                        yield from iter_component_schemas(item_schema)
            # object schema type
            case {"properties": props}:
                yield from iter_component_schemas(props)
            # any other actual schema that has a 'type' key. At this point, it cannot
            # be a container, as these have been handled before.
            case {"type": Sequence()}:
                yield schema
            case {"oneOf": nested} | {"allOf": nested} | {"anyOf": nested}:
                yield from iter_component_schemas(nested)
            case [*nested]:
                for child in nested:
                    yield from iter_component_schemas(child)
            case MutableMapping():
                for nested in schema.values():
                    yield from iter_component_schemas(nested)

    schemas_iterator = itertools.chain(
        iter_parameter_schemas(),
        iter_component_schemas(result.get("components", {}).get("schemas", {})),
    )

    # find all string (with format uri) properties that have an invalid default
    for schema in schemas_iterator:
        # only consider string types - the first match skips over everything that's
        # not a string type schema
        match schema:
            case {"type": "string"}:
                pass
            case {"type": [*types]} if "string" in types:
                pass
            case _:
                continue

        # this actual processes the default value for URI schemas
        match schema:
            case {"format": "uri", "default": ""}:
                del schema["default"]
            case _:
                continue

    return result



from typing import Any

from django import template
from django.utils.html import escape, format_html
from django.utils.safestring import mark_safe

from maykin_common.settings import get_setting

register = template.Library()


@register.inclusion_tag("maykin_common/includes/version_info.html")
def show_version_info():
    """
    Template that displays version info.
    """
    return {
        "RELEASE": get_setting("RELEASE"),
        "GIT_SHA": get_setting("GIT_SHA"),
    }


@register.simple_tag(takes_context=True)
def show_environment_info(context: dict[str, Any]) -> str:
    """
    Template that show the current ENVIRONMENT to an authenticated user.

    Returns an empty string if SHOW_ENVIRONMENT is set to `False`
    """
    if not get_setting("SHOW_ENVIRONMENT"):
        return ""
    if (user := context.get("user")) is None or not user.is_authenticated:
        return ""

    style_tokens = {
        "background-color": get_setting("ENVIRONMENT_BACKGROUND_COLOR"),
        "color": get_setting("ENVIRONMENT_FOREGROUND_COLOR"),
    }
    _inline_style_bits = [
        f"--admin-env-info-{key}: {value}".format(key=key, value=escape(value))
        for key, value in style_tokens.items()
    ]
    return format_html(
        """<div class="env-info" style="{style}">{label}</div>""",
        label=get_setting("ENVIRONMENT_LABEL"),
        style=mark_safe("; ".join(_inline_style_bits)),
    )



from django.db import migrations, router

RESET_SQL = """
    SELECT 'SELECT SETVAL(' ||
       quote_literal(quote_ident(PGT.schemaname) || '.' || quote_ident(S.relname)) ||
       ', COALESCE(MAX(' ||quote_ident(C.attname)|| '), 1) ) FROM ' ||
       quote_ident(PGT.schemaname)|| '.'||quote_ident(T.relname)|| ';'
    FROM pg_class AS S,
         pg_depend AS D,
         pg_class AS T,
         pg_attribute AS C,
         pg_tables AS PGT
    WHERE S.relkind = 'S'
        AND S.oid = D.objid
        AND D.refobjid = T.oid
        AND D.refobjid = C.attrelid
        AND D.refobjsubid = C.attnum
        AND T.relname = PGT.tablename
    ORDER BY S.relname;
    """


class ResetSequences(migrations.RunSQL):
    """
    Run the reset_sequences SQL.

    Resetting the Postgres sequences makes sure you don't get Integrity Errors
    when creating objects because of failing unique constraints. This happens
    when you create records explicitly with PKs, bypassing the database PK
    generation. Resetting the sequences makes sure the sequences are aware of
    the used PKs.

    Usage:

        >>> from maykin_common.migration_operations import ResetSequences
        >>> class Migration(migrations.Migration):
        ...     dependencies = (...)
        ...     operations = [
        ...         ResetSequences(),
        ...     ]
    """

    def __init__(self, *args, **kwargs):
        super().__init__("", *args, **kwargs)

        self.reverse_sql = None

    def database_forwards(self, app_label, schema_editor, from_state, to_state) -> None:
        if router.allow_migrate(
            schema_editor.connection.alias, app_label, **self.hints
        ):
            with schema_editor.connection.cursor() as cursor:
                cursor.execute(RESET_SQL)
                rows = cursor.fetchall()

            sql = "\n".join(x[0] for x in rows)

            self._run_sql(schema_editor, sql)  # pyright:ignore[reportAttributeAccessIssue]

    def database_backwards(self, *args, **kwargs) -> None:
        pass



"""
Provides Mixins and TestCases for snapshot testing outgoing HTTP requests
with VCR_.

Depends on vcrpy_ and requests_, but is also usable with `other HTTP
libraries`_ than requests.

.. note::
    The sub-dependencies will come automatically with the ``vcr`` extras:
    ``maykin-common[vcr]``. Because this is a test tool, only include it in your CI
    and dev dependencies. Don't include the ``vcr`` extra in the base dependencies
    of your project.


.. _VCR: https://vcrpy.readthedocs.io
.. _vcrpy: https://pypi.org/project/vcrpy/
.. _requests: https://pypi.org/project/requests/
.. _other HTTP libraries: https://vcrpy.readthedocs.io/en/latest/installation.html#compatibility
"""

import inspect
import os
from collections.abc import Callable
from contextlib import AbstractContextManager
from pathlib import Path
from typing import Any, Protocol, override

from django.test import SimpleTestCase, TestCase, TransactionTestCase, tag

import requests.exceptions
from vcr.cassette import Cassette
from vcr.config import RecordMode
from vcr.request import Request
from vcr.unittest import VCR, VCRMixin as _VCRMixin

__all__ = [
    "SimpleVCRTestCase",
    "TransactionVCRTestCase",
    "VCRMixin",
    "VCRTestCase",
]
type _VCRBRRHook = Callable[[Request], Request | None]
"""VCR before_record_request hook
May mutate Request and return it, or cancel recording by returning None"""


class _VCRTestCase(Protocol):
    "The interface VCRMixin depends on"

    _testMethodName: str

    def _get_cassette_name(self) -> str: ...
    def _get_vcr(self, **kwargs) -> VCR: ...
    # TypedDict is almost as cursed as Protocol :(
    def _get_vcr_kwargs(self, **kwargs) -> dict[str, Any]: ...


class VCRMixin(_VCRMixin):
    """
    Mixin to use VCR cassettes to record HTTP requests/responses.
    """

    vcr_enabled: bool
    """
    Easy toggle to temporarily turn vcr *off* during development or debugging.
    So `True` does *not* enable recording "episodes" that would otherwise not
    be recorded.
    """

    VCR_RECORD_MODE: RecordMode = RecordMode(
        os.environ.get("VCR_RECORD_MODE", RecordMode.NONE)
    )
    """
    Defaults to `VCR_RECORD_MODE` env variable or `RecordMode.NONE`.
    To (re-)record throw away the cassettes and set to `RecordMode.ONCE`
    """

    VCR_TEST_FILES: Path | None = None
    """
    Cassettes will be stored in

      ``VCR_TEST_FILES``/vcr_cassettes/`{test class name}`/`{test method name}.yaml`

    If left `None`, a ``files`` directory at the same level as the test class file
    will be used.
    """

    @override
    def _get_cassette_library_dir(self):
        test_files = (
            self.VCR_TEST_FILES
            or Path(inspect.getfile(self.__class__)).parent / "files"
        )
        test_files.mkdir(exist_ok=True)
        return str(test_files / "vcr_cassettes" / self.__class__.__qualname__)

    @override
    def _get_cassette_name(self: _VCRTestCase):
        """Return the filename for cassette

        Default VCR behaviour puts class name in the cassettename
        we put them in a directory.
        """
        return f"{self._testMethodName}.yaml"

    @override
    def _get_vcr_kwargs(self, **kwargs):
        return {
            "record_mode": self.VCR_RECORD_MODE,
            # Decompress for human readable cassette diffs when re-recoding
            "decode_compressed_response": True,
        } | super()._get_vcr_kwargs(**kwargs)

    def vcr_raises(
        self: _VCRTestCase,
        exception: Callable[[], Exception] = requests.exceptions.RequestException,
    ) -> AbstractContextManager[Cassette]:
        """Simulate occurrence of an error during HTTP request.

        Example:

        .. code-block::

            from requests.exceptions import SSLError, Timeout

            # sometimes people let certificates expire
            with self.vcr_raises(SSLError):
                response = function_under_test_that_uses_requests()

            # or services/connections are down
            with self.vcr_raises(Timeout):
                response = function_under_test_that_uses_requests()

        .. note::
           Instead of performing and recording a request, this raises an exception.
           So there will be no request nor cassette!

        """
        # TODO: decouple exception from requests with generic Timeout/SSLError/etc that
        # inherit from all semantically equal exceptions thrown by better libraries
        # than requests. A client can then change its implementation without a need to
        # change the tests (iff it doesn't do anything with the Error arguments).

        kwargs = self._get_vcr_kwargs()
        hook: _VCRBRRHook = kwargs.get("before_record_request") or (lambda _: None)

        def raise_exception(request):
            # perform configured hook first
            hook(request)
            raise exception()

        clean_vcr = self._get_vcr(**kwargs | {"before_record_request": raise_exception})
        return clean_vcr.use_cassette(self._get_cassette_name())


@tag("vcr")
class SimpleVCRTestCase(VCRMixin, SimpleTestCase):
    """A Django ``SimpleTestCase`` with the ``VCRMixin`` and a ``vcr`` tag.

    Use this for testing "client code".
    """


@tag("vcr")
class VCRTestCase(VCRMixin, TestCase):
    """A Django ``TestCase`` with the ``VCRMixin`` and a ``vcr`` tag.

    Use this if your `Model` may do HTTP requests.
    """


@tag("vcr")
class TransactionVCRTestCase(VCRMixin, TransactionTestCase):
    """A Django ``TestCase`` with the ``VCRMixin`` and a ``vcr`` tag.

    Use this only if you really need it. i.e when you're actually testing the
    transactional behaviour of your app.
    """



"""
Utilities for PDF rendering from HTML using WeasyPrint.

Note that you need to add https://pypi.org/project/weasyprint/ to your dependencies
if you want to make use of HTML-to-PDF rendering. This is not included by default as
it's quite heavy and requires OS-level dependencies.

This module exposes the public function :func:`render_to_pdf` which renders a template
with a context into a PDF document (bytes output). You can use "external" stylesheets
in these templates, and they will be resolved through django's staticfiles machinery
by the custom :class:`UrlFetcher`.

Depends on ``weasyprint``.
"""

import functools
import logging
import mimetypes
from collections.abc import Mapping
from io import BytesIO
from pathlib import PurePosixPath
from typing import NotRequired, TypedDict
from urllib.parse import ParseResult, urlparse

from django.conf import settings
from django.contrib.staticfiles import finders
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.files.storage import FileSystemStorage, default_storage
from django.core.signals import setting_changed
from django.dispatch import receiver
from django.template.loader import render_to_string
from django.utils.module_loading import import_string

import weasyprint

from maykin_common.settings import get_setting

logger = logging.getLogger(__name__)

__all__ = ["render_to_pdf", "render_template_to_pdf"]


def get_base_url() -> str:
    """
    Get the base URL where the project is served.
    """

    if pdf_base_url_function := get_setting("PDF_BASE_URL_FUNCTION"):
        return import_string(pdf_base_url_function)()
    raise NotImplementedError("You must implement 'get_base_url'.")


def _ensure_fully_qualified_url(url: str, base: ParseResult) -> ParseResult:
    """
    Ensure the passed in URL is fully qualified.

    If the URL does not have a network location, we take the protocol and netloc from
    the provided base URL to make it fully qualified. This assumes no netloc implies
    no protocol.
    """
    parsed_url = urlparse(url)
    match parsed_url:
        case ParseResult(scheme=scheme, netloc=netloc) if scheme and netloc:
            return parsed_url
        case _:
            # it is documented as public API!
            return parsed_url._replace(scheme=base.scheme, netloc=base.netloc)


@functools.cache
def _get_candidate_storages() -> Mapping[ParseResult, FileSystemStorage]:
    """
    Introspect settings and determine which storages can serve static assets.

    We can only consider storages that inherit from :class:`FileSystemStorage` for
    optimized asset serving. The goal of this module is to avoid network round-trips to
    our own ``MEDIA_ROOT`` or ``STATIC_ROOT``.
    """
    base_url = urlparse(get_base_url())
    candidates: dict[ParseResult, FileSystemStorage] = {}

    # check staticfiles app
    if isinstance(staticfiles_storage, FileSystemStorage):
        static_url = _ensure_fully_qualified_url(settings.STATIC_URL, base=base_url)
        candidates[static_url] = staticfiles_storage

    # check media root
    if isinstance(default_storage, FileSystemStorage):
        media_url = _ensure_fully_qualified_url(settings.MEDIA_URL, base=base_url)
        candidates[media_url] = default_storage

    return candidates


@receiver(setting_changed, dispatch_uid="maykin_common.pdf._reset_storages")
def _reset_storages(sender, setting: str, **kwargs):
    # mostly for tests, settings *should* not change in production code
    match setting:
        case "STATIC_ROOT" | "MEDIA_ROOT" | "STORAGES" | "PDF_BASE_URL_FUNCTION":
            _get_candidate_storages.cache_clear()
        case _:  # pragma: no cover
            pass


class UrlFetcherResult(TypedDict):
    mime_type: str | None
    encoding: str | None
    redirected_url: str
    filename: str
    file_obj: NotRequired[BytesIO]
    string: NotRequired[bytes]


class UrlFetcher:
    """
    URL fetcher that skips the network for /static/* and /media/* files.
    """

    def __call__(self, url: str) -> UrlFetcherResult:
        """
        Check if the URL matches one of our candidates and use it if there's a match.

        Matching is done on the URLs of the storages and the requested asset. If the
        prefix matches, look up the relative asset path in the storage and serve it
        if it's found. If not, defer to the default URL fetcher of WeasyPrint.
        """
        parsed_url = urlparse(url)
        assert parsed_url.netloc and parsed_url.netloc, "Expected fully qualified URL"

        # Try candidates, respecting the order of the candidate configuration.
        for base, storage in _get_candidate_storages().items():
            base_url = base.geturl()
            # Skip to the next candidate if the URLs don't share a prefix.
            if not url.startswith(base_url):
                continue

            # get the relative path to lookup in the storage to obtain an absolute path
            rel_path = PurePosixPath(parsed_url.path).relative_to(base.path)
            rel_path_str = str(rel_path)

            absolute_path: str | None = None
            if storage.exists(rel_path_str):
                absolute_path = storage.path(rel_path_str)
            elif settings.DEBUG and storage is staticfiles_storage:
                # use finders so that it works in dev too, we already check that it's
                # using filesystem storage earlier
                absolute_path = finders.find(rel_path_str)

            # we bail out, since we hit a storage that matches the URL prefix. Other
            # candidates will not have match either due to their different URL prefixes.
            if absolute_path is None:
                logger.error(
                    "path_resolution_failed",
                    extra={
                        "path": rel_path_str,
                        "storage": storage,
                    },
                )
                return weasyprint.default_url_fetcher(url)  # pyright:ignore[reportReturnType]

            content_type, encoding = mimetypes.guess_type(absolute_path)
            result: UrlFetcherResult = {
                "mime_type": content_type,
                "encoding": encoding,
                "redirected_url": url,
                "filename": rel_path.parts[-1],
            }
            with open(absolute_path, "rb") as f:
                result["file_obj"] = BytesIO(f.read())
            return result

        else:
            # all candidates were tried, none were a match -> defer to the weasyprint
            # default
            return weasyprint.default_url_fetcher(url)  # pyright:ignore[reportReturnType]


def render_to_pdf(html: str, variant: str | None = "pdf/ua-1") -> tuple[str, bytes]:
    """
    Render the provided HTML to PDF.

    The default ``variant`` generates accessible PDFs. Technically it's still an
    experimental feature in WeasyPrint, so if it's causing issues, you can pass
    ``variant=None`` instead.
    """
    html_object = weasyprint.HTML(
        string=html,
        url_fetcher=UrlFetcher(),
        base_url=get_base_url(),
    )
    pdf = html_object.write_pdf(pdf_variant=variant)
    assert isinstance(pdf, bytes)
    return html, pdf


def render_template_to_pdf(
    template_name: str,
    context: dict[str, object],
    variant: str | None = "pdf/ua-1",
) -> tuple[str, bytes]:
    """
    Render a (HTML) template to PDF with the given context.
    """
    rendered_html = render_to_string(template_name, context=context)
    return render_to_pdf(rendered_html, variant=variant)



"""
Open Telemetry integration in a Django project.

The primary export of this module is :func:`setup_otel`, which is intended to be called
from the ``setup_env`` function in projects.

The Python OpenTelemetry SDK supports the standardized environment variables,
see the `environment variables reference`_.

We define one custom environment variable ``_OTEL_ENABLE_CONTAINER_RESOURCE_DETECTOR``
to opt-in to container resource detection when the application is being deployed on a
non-kubernetes container runtime (such as Docker or Podman).

.. _`environment variables reference`: https://opentelemetry.io/docs/specs/otel/configuration/sdk-environment-variables/
"""

import os
from typing import Literal, assert_never
from uuid import uuid4

from django.core.exceptions import ImproperlyConfigured

from opentelemetry import metrics, trace
from opentelemetry.instrumentation.django import DjangoInstrumentor
from opentelemetry.sdk.environment_variables import OTEL_EXPORTER_OTLP_PROTOCOL
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import (
    DEPLOYMENT_ENVIRONMENT,
    SERVICE_INSTANCE_ID,
    SERVICE_VERSION,
    Resource,
    get_aggregated_resources,
)
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from maykin_common.settings import get_setting

# the uwsgi module is special - it's only available when the python code is loaded
# through uwsgi. With regular ``manage.py`` usage, it does not exist.
try:
    import uwsgi  # pyright: ignore[reportMissingModuleSource] uwsgi magic...
except ImportError:
    uwsgi = None

__all__ = [
    "setup_otel",
]

# The Python SDK only supports gRPC and HTTP/protobuf. HTTP/json is not supported.
type ExportProtocol = Literal["grpc", "http/protobuf"]

DEFAULT_PROTOCOL: ExportProtocol = "grpc"


def setup_otel() -> None:
    """
    Initialize the Open Telemetry machinery.

    This is an application-global hook to configure the Open Telemetry machinery:

    * set up the metadata about the resource being monitored
    * initialize the metrics processor and exporter
    * initialize the traces processor and exporter
    * instrument the python code so that metrics and traces are captured

    The exports are responsible for shipping the telemetry data to an endpoint that
    supports OpenTelemetryProtocol (OTLP), either over gRPC or HTTP/protobuf protocols.
    Often, this will be an OpenTelemetry Collector running somewhere.

    Part of the SDK initialization process is starting a background thread to
    periodically ship the telemetry to the configured endpoint.
    """
    # set up instrumenters that (usually) monkeypatch modules or inject the right
    # wrappers/middleware etc.

    # the instrumentor is a singleton, so it's effectively global
    instrumentor = DjangoInstrumentor()
    if not instrumentor.is_instrumented_by_opentelemetry:
        instrumentor.instrument()

    # in a uwsgi worker, defer the otel initialization until after the processes have
    # forked
    if uwsgi is not None:  # pragma: no cover - can't be tested outside of uwsgi
        from uwsgidecorators import postfork

        postfork(_setup_otel)
    else:
        _setup_otel()


def _setup_otel() -> None:
    """
    Helper function for the actual initialization.

    This helpers makes it possible to actually properly initialize in a non-uwsgi
    context, while deferring the initialization until post-fork in a uwsgi context.
    """
    _already_initialized = isinstance(trace.get_tracer_provider(), TracerProvider)
    if _already_initialized:
        return

    if "OTEL_SERVICE_NAME" not in os.environ:
        raise ImproperlyConfigured(
            "You must define the 'OTEL_SERVICE_NAME' environment variable."
        )
    # the service name now is guaranteed to be set through envvars
    resource = Resource.create(
        attributes={
            SERVICE_VERSION: get_setting("RELEASE") or "",
            SERVICE_INSTANCE_ID: str(uuid4()),
            DEPLOYMENT_ENVIRONMENT: get_setting("ENVIRONMENT") or "",
        }
    )
    resource = aggregate_resource(resource)

    OTLPMetricExporter, OTLPSpanExporter = load_exporters()

    tracer_provider = TracerProvider(resource=resource)
    processor = BatchSpanProcessor(OTLPSpanExporter())
    tracer_provider.add_span_processor(processor)
    trace.set_tracer_provider(tracer_provider)

    reader = PeriodicExportingMetricReader(OTLPMetricExporter())
    meter_provider = MeterProvider(resource=resource, metric_readers=[reader])
    metrics.set_meter_provider(meter_provider)


def load_exporters():
    # TODO: replace with `config` helper once it's added to this library
    protocol: ExportProtocol = os.getenv(  # pyright: ignore[reportAssignmentType]
        OTEL_EXPORTER_OTLP_PROTOCOL, default=DEFAULT_PROTOCOL
    )
    match protocol:
        case "grpc":
            from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import (
                OTLPMetricExporter,
            )
            from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
                OTLPSpanExporter,
            )

            return (OTLPMetricExporter, OTLPSpanExporter)
        case "http/protobuf":
            from opentelemetry.exporter.otlp.proto.http.metric_exporter import (
                OTLPMetricExporter,
            )
            from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
                OTLPSpanExporter,
            )

            return (OTLPMetricExporter, OTLPSpanExporter)
        case _:
            assert_never(protocol)


def aggregate_resource(resource: Resource) -> Resource:
    # TODO: replace with `config` helper once it's added to this library
    _enable_resource_detector = (
        os.getenv("_OTEL_ENABLE_CONTAINER_RESOURCE_DETECTOR", default="false")
        .lower()
        .strip()
        == "true"
    )
    if not _enable_resource_detector:
        return resource

    from opentelemetry.resource.detector.containerid import ContainerResourceDetector

    return get_aggregated_resources(
        detectors=[ContainerResourceDetector()], initial_resource=resource
    )



from typing import Literal

from django.conf import settings

GOOGLE_ANALYTICS_ID = None
"""
Google analytics id added to the request context.
"""

SHOW_ENVIRONMENT = False
"""
Enable or disable the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_BACKGROUND_COLOR = "orange"
"""
The background color of the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_FOREGROUND_COLOR = "black"
"""
The foreground color of the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_LABEL = None
"""
The textual content of the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT = ""
"""
The deployment environment label, e.g. 'staging' or 'prod'.

This setting is included in the OpenTelemetry resource attributes.
"""

RELEASE = None
"""
The release version shown in the ``{% version_info %}`` template tag.

This setting is included in the OpenTelemetry resource attributes.
"""

GIT_SHA = None
"""
The commit hash shown in the ``{% version_info %}`` template tag.
"""

PDF_BASE_URL_FUNCTION = None
"""
Function that returns the base url needed to download/resolve custom fonts and/or any
image URLs included in the document to render.

Required for the :ref:`quickstart_pdf` extra.
"""

type SettingName = Literal[
    "GOOGLE_ANALYTICS_ID",
    "ENVIRONMENT",
    "SHOW_ENVIRONMENT",
    "ENVIRONMENT_BACKGROUND_COLOR",
    "ENVIRONMENT_FOREGROUND_COLOR",
    "ENVIRONMENT_LABEL",
    "RELEASE",
    "GIT_SHA",
    "PDF_BASE_URL_FUNCTION",
]


def get_setting(name: SettingName):
    default = globals()[name]
    return getattr(settings, name, default)
